;;; smart-find.el: Find a file in any of a predetermined set of directories
;;; Copyright (c) 1990 by Wayne Mesard.
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 1, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; The GNU General Public License is available by anonymouse ftp from
;;; prep.ai.mit.edu in pub/gnu/COPYING.  Alternately, you can write to
;;; the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
;;; USA.
;;--------------------------------------------------------------------

;;; DESCRIPTION:
;;   Allows you to quickly find a file in one of a prespecified set of
;;   directories, without requiring you to think about (or specify)
;;   which directory the file is in.
;;   
;;   Does filename completion and possible match listings, just like
;;   regular old find-file.
;;
;;   By default SMART-FIND-FILE is bound to Control-C Control-F.  To prevent
;;   this, add "(setq dont-bind-my-keys t)" to the top of your ~/.emacs file.
;;   
;;   To specify the search path SMART-FIND-FILE uses, set the variable 
;;   SMART-FIND-FILE-PATH.  E.g., put something like this in your ~/.emacs:
;;
;;     (setq smart-find-file-path
;;        '("~" "/usr/include" "/usr/include/sys" "$myworkdir"))
;;
;;   By default, it looks in your home directory.

;;; BUGS
;;   - Does the wrong thing on minibuffer-complete (TAB): e.g.: DIR1 contains 
;;     file "foo", DIR2 contains file "fab".  Type "f[TAB]", and it completes
;;     to "foo" (it should make a completion list).
;;   - Can't select files from subdirectoies on the path.

;;; HISTORY:
;;  1.1 wmesard - Jun 23, 1991: 
;;	 Fixed bug where "foo" couldn't be loaded if "foo.c" existed.
;;       Tab now says the right thing "Sole completion", etc.
;;       Should work on VMS too.
;;  1.0 wmesard - Sep 27, 1990: Created.

;;; 
;;; KEY BINDINGS
;;; 

(if (not (and (boundp 'dont-bind-my-keys) dont-bind-my-keys))
    (global-set-key "\C-c\C-f" 'smart-find-file))

;;; 
;;; USER OPTIONS
;;; 

(defvar smart-find-file-path '("~")
  "*Which directories smart-find-file should examine.")

;;; 
;;; SYSTEM VARIABLES
;;; 

(defvar smart-find-file-match)

;;; 
;;; COMMANDS
;;; 

(defun smart-find-file (filename)
  (interactive (list (completing-read "Smart find file: "
				      'smart-find-file-internal
				      nil t nil)))
  (find-file smart-find-file-match))

;;; 
;;; PRIVATE FUNCTIONS
;;; 

;; Look along the path, return the first match you find.
;; Partial matches are allowed and it will return 
;; the longest string common to all matches in the first 
;; matching directory.

(defun smart-find-file-lookup (fn)
  (let ((path smart-find-file-path)
	(result nil))
    (while (and path 
		(not (setq result 
			   (file-name-completion
			    fn (substitute-in-file-name (car path))))
		     ))
      (setq path (cdr path)))
    (setq smart-find-file-match 
	  (if (eq result t)
	      (substitute-in-file-name (expand-file-name fn (car path)))
	    result))
    result))



(defun smart-find-file-all (str)
  (apply 'append 
	 (mapcar '(lambda (x) 
		    (file-name-all-completions 
		     str (substitute-in-file-name x)))
		 smart-find-file-path)
	 ))


;; Called by completing-read to Complete, List matches or Verify
;; a candidate filename, when the user hits Tab, Question mark,
;; or Return, respectively.
;;   Returns: T or NIL on Verify.
;;            the filename or partial filename on Complete.
;;            A list of matches from all directories in the path on List.

(defun smart-find-file-internal (str ignore action)
  "Internal subroutine for smart-find-file.  Do not call this."
  ;; ACTION is nil for complete, t for return list of completions,
  ;; lambda for verify final value
  (cond ((null action)			; ACTION = Complete
	 (smart-find-file-lookup str))
	((eq action 'lambda)		; ACTION = Verify
	 (let ((path smart-find-file-path)
	       result)
	   (while
	       (and path 
		    (not (file-exists-p
			  (setq result (substitute-in-file-name 
					(expand-file-name str (car path)))))))
	     (setq path (cdr path)))
	   (if path
	       (progn
		 (setq smart-find-file-match result)
		 t))
	   ))
	(t				; ACTION = List matches
	 (smart-find-file-all str))
	))
